use crate::{Connection, Error};
use nostr_rs_storage::RocksDb;
use nostr_rs_types::{
    relayer,
    types::{Event, Filter, SubscriptionId},
    Request, Response,
};
use parking_lot::{RwLock, RwLockReadGuard};
use std::{collections::HashMap, ops::Deref, sync::Arc};
use tokio::{
    net::TcpStream,
    sync::mpsc::{channel, Receiver, Sender},
};

#[derive(Clone, Debug, Default, Eq, PartialEq, Hash)]
pub struct SubscriptionType {
    public_key: Option<Vec<u8>>,
    id: Option<Vec<u8>>,
    kind: Option<u32>,
}

type SubId = u128;

type Subscriptions = HashMap<SubId, (SubscriptionId, Sender<Response>)>;

pub struct Relayer {
    storage: RocksDb,
    /// Keeps a map between the internal subscription ID and the subscription
    /// type. One subscription ID may have multiple subscription types.
    ///
    /// Each connection keeps a list of the subscription ID provided by the user
    /// (String) and the internal, globally recognized subscription ID which is
    /// internal (SubId)
    subscriptions_ids_index: RwLock<HashMap<SubId, Vec<SubscriptionType>>>,
    /// Each subscription type that is active has a list of subscriptions.
    ///
    /// A single REQ can be subscribed to multiple subscription types, specially
    /// when it is translated in OR filters. It is designed this way to allow a
    /// fast iteration and match quickly filters.
    subscriptions: RwLock<HashMap<SubscriptionType, RwLock<Subscriptions>>>,
    clients: RwLock<HashMap<u128, Connection>>,
    sender: Sender<(u128, Request)>,
}

impl Relayer {
    pub fn new(storage: RocksDb) -> (Arc<Self>, Receiver<(u128, Request)>) {
        let (sender, receiver) = channel(100_000);
        (
            Arc::new(Self {
                storage,
                subscriptions: RwLock::new(HashMap::new()),
                subscriptions_ids_index: RwLock::new(HashMap::new()),
                clients: RwLock::new(HashMap::new()),
                sender,
            }),
            receiver,
        )
    }

    /// Returns a reference to the internal database
    pub fn get_db(&self) -> &RocksDb {
        &self.storage
    }

    pub async fn add_connection(&self, stream: TcpStream) -> Result<(), Error> {
        let client = Connection::new(self.sender.clone(), stream).await?;
        let mut clients = self.clients.write();
        clients.insert(client.conn_id, client);

        Ok(())
    }

    pub async fn recv(
        &self,
        receiver: &mut Receiver<(u128, Request)>,
    ) -> Result<Option<Request>, Error> {
        let (conn_id, request) = if let Some(request) = receiver.recv().await {
            request
        } else {
            return Ok(None);
        };

        let connections = self.clients.read();
        let connection = connections
            .get(&conn_id)
            .ok_or(Error::UnknownConnection(conn_id))?;

        match &request {
            Request::Event(event) => {
                self.store_and_broadcast_local_event(event.deref());
            }
            Request::Request(request) => {
                for filter in request.filters.clone().into_iter() {
                    // Create subscription
                    let (sub_id, receiver) =
                        connection.create_subscription(request.subscription_id.deref().to_owned());
                    let mut sub_index = self.subscriptions_ids_index.write();
                    let mut subscriptions = self.subscriptions.write();
                    if let Some(prev_subs) = sub_index.remove(&sub_id) {
                        // remove any previous subscriptions
                        prev_subs.iter().for_each(|index| {
                            if let Some(subscriptions) = subscriptions.get_mut(&index) {
                                subscriptions.write().remove(&sub_id);
                            }
                        });
                    }
                    sub_index.insert(
                        sub_id,
                        Self::get_indexes_from_filter(&filter)
                            .into_iter()
                            .map(|index| {
                                subscriptions
                                    .entry(index.clone())
                                    .or_insert_with(|| RwLock::new(HashMap::new()))
                                    .write()
                                    .insert(
                                        sub_id,
                                        (request.subscription_id.clone(), receiver.clone()),
                                    );
                                index
                            })
                            .collect::<Vec<_>>(),
                    );

                    drop(subscriptions);
                    drop(sub_index);

                    // Sent all events that match the filter that are stored in our database
                    self.storage
                        .get_by_filter(filter)?
                        .into_iter()
                        .for_each(|event| {
                            let _ = connection.send(
                                relayer::Event {
                                    subscription_id: request.subscription_id.clone(),
                                    event,
                                }
                                .into(),
                            );
                        });
                }
                let _ = connection
                    .send(relayer::EndOfStoredEvents(request.subscription_id.clone()).into());
            }
            Request::Close(close) => {
                connection.get_subscription_id(&*close.0).map(|id| {
                    let mut subscriptions = self.subscriptions_ids_index.write();
                    subscriptions.remove(&id).map(|indexes| {
                        let mut subscriptions = self.subscriptions.write();
                        indexes.iter().for_each(|index| {
                            if let Some(subscriptions) = subscriptions.get_mut(&index) {
                                subscriptions.write().remove(&id);
                            }
                        });
                    });
                });
            }
        };

        Ok(Some(request))
    }

    #[inline]
    fn get_indexes_from_filter(filter: &Filter) -> Vec<SubscriptionType> {
        let public_keys = if filter.references_to_public_key.is_empty() {
            vec![None]
        } else {
            filter
                .references_to_public_key
                .iter()
                .map(|public_key| Some((*public_key).to_vec()))
                .collect()
        };
        let id = if filter.references_to_event.is_empty() {
            vec![None]
        } else {
            filter
                .references_to_event
                .iter()
                .map(|id| Some((*id).to_vec()))
                .collect()
        };
        let kind = if filter.kinds.is_empty() {
            vec![None]
        } else {
            filter
                .kinds
                .iter()
                .map(|kind| Some((*kind).into()))
                .collect()
        };
        let mut subs = vec![];

        for public_key in public_keys.iter() {
            for id in id.iter() {
                for kind in kind.iter() {
                    subs.push(SubscriptionType {
                        public_key: public_key.clone(),
                        id: id.clone(),
                        kind: *kind,
                    });
                }
            }
        }

        subs
    }

    #[inline]
    fn get_possible_listeners_from_event(event: &Event) -> Vec<SubscriptionType> {
        let kind = event.kind().into();
        let public_keys = [None, Some(event.author().as_ref().to_vec())];
        let id = [None, Some(event.id.as_ref().to_vec())];
        let kind = [None, Some(kind)];
        let mut subs = vec![];

        for public_key in public_keys.iter() {
            for id in id.iter() {
                for kind in kind.iter() {
                    subs.push(SubscriptionType {
                        public_key: public_key.clone(),
                        id: id.clone(),
                        kind: *kind,
                    });
                }
            }
        }

        subs
    }

    #[inline]
    fn broadcast_to_subscribers(subscriptions: RwLockReadGuard<Subscriptions>, event: &Event) {
        for (_, receiver) in subscriptions.iter() {
            let _ = receiver.1.try_send(
                relayer::Event {
                    subscription_id: receiver.0.clone(),
                    event: event.clone(),
                }
                .into(),
            );
        }
    }

    #[inline]
    pub fn store_and_broadcast_local_event(&self, event: &Event) {
        let _ = self.storage.store_local_event(event);
        let subscriptions = self.subscriptions.read();

        for subscription_type in Self::get_possible_listeners_from_event(event) {
            if let Some(subscribers) = subscriptions.get(&subscription_type) {
                Self::broadcast_to_subscribers(subscribers.read(), event);
            }
        }
    }

    #[inline]
    pub fn store_and_broadcast(&self, event: &Event) {
        let _ = self.storage.store(event);
        let subscriptions = self.subscriptions.read();

        for subscription_type in Self::get_possible_listeners_from_event(event) {
            if let Some(subscribers) = subscriptions.get(&subscription_type) {
                Self::broadcast_to_subscribers(subscribers.read(), event);
            }
        }
    }
}
